/**************************************************************************
 *
 * This is jnet, the portable pocketknife.
 * 
 *****************************************************************************/

#include <sys/param.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/stat.h>
#include <errno.h>
#include <stropts.h>
#include <poll.h>
#include <netdb.h>
#include <stdio.h>
#include <stdarg.h>
#include <fcntl.h>
#include <pwd.h>
#include <sys/wait.h>
#include <signal.h>
#include <unistd.h>
#include <termios.h>
#include <setjmp.h>
#include <string.h>
#include <stdlib.h>

void usage();

#ifdef STRERROR
extern char *sys_errlist[];
#define strerror(x) (sys_errlist[x])
#endif

/*****************************************************************************
 *
 * Simple Stuff
 *
 *****************************************************************************/

char *rootify(char *path)
{
  if (path[0]==0) return "/";
}

void jseed()
{
  srand48(time(0));
}

int jrand()
{
  return lrand48();
}
 
int filetime(char *path)
{
  struct stat info;
  int ok = stat(path, &info);
  if (ok<0) {
    fprintf(stderr,"%s: %s\n", path, strerror(errno));
    exit(2);
  }
  return info.st_mtime;
}

double now()
{
  double res;
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return ((double)tv.tv_sec) + (((double)tv.tv_usec)*1.0E-6);
}

FILE *efopen(char *fn, char *mode)
{
  FILE *res = fopen(fn, mode);
  if (res==0) {
    fprintf(stderr,"Cannot open %s for %s\n", fn, mode);
    exit(1);
  }
  return res;
}

int writeall(int fd, char *buff, int len)
{
  int nwrote;
  while (len > 0) {
    nwrote = write(fd, buff, len);
    if (nwrote < 0) { return -1; }
    len -= nwrote; buff += nwrote;
  }
  return 0;
}

void ewriteall(int fd, char *buff, int len)
{
  int ok = writeall(fd, buff, len);
  if (ok<0) { perror("write"); exit(1); }
}

FILE *tfopen(char *path)
{
  char fn[MAXPATHLEN]; int fd;
  if (path==0) return 0;
  if ((path[0]=='-')&&(path[1]==0)) return stdout;
  sprintf(fn,"%s~",path);
  fd = open(fn, O_WRONLY | O_CREAT | O_TRUNC, 0644);
  if (fd < 0) {
    fprintf(stderr,"Cannot open %s: %s\n", path, strerror(errno));
    exit(1);
  }
  return fdopen(fd, "r+");
}

void tfclose(FILE *f, char *path)
{
  char fn[MAXPATHLEN];
  if (path==0) return;
  fclose(f);
  if ((path[0]=='-')&&(path[1]==0)) return;
  sprintf(fn,"%s~",path);
  rename(fn, path);
}

void relay(int src0, int sink0, int src1, int sink1)
{
  char buffer[1024]; struct pollfd fds[2];
  int shutdowns = 0; int ok, nread;
  fds[0].fd = src0; fds[0].events = POLLIN; 
  fds[1].fd = src1; fds[1].events = POLLIN;
  while (shutdowns < 2) {
    fds[1].revents = 0;
    fds[0].revents = 0;
    ok = poll(fds, 2, -1);
    if (fds[0].revents) {
      if (fds[0].revents != POLLIN) exit(0);
      nread = read(src0, buffer, 1024);
      if (nread<0) exit(0);
      if (nread>0) ewriteall(sink0, buffer, nread);
      if (nread==0) { shutdowns++; shutdown(sink0, 1); }
    }
    if (fds[1].revents) {
      if (fds[1].revents != POLLIN) exit(0);
      nread = read(src1, buffer, 1024);
      if (nread<0) exit(0);
      if (nread>0) ewriteall(sink1, buffer, nread);
      if (nread==0) { shutdowns++; shutdown(sink1, 1); }
    }
  }
  close(src0);
  close(src1);
  close(sink0);
  close(sink1);
}

#define eqs(s1,s2) (strcmp((s1), (s2))==0)

/*****************************************************************************
 *
 * Wildcard matching
 *
 ****************************************************************************/

int string_match(char *str, char *pat)
{
  char c1, c2;
top:
  if (pat[0]==0) return (str[0]==0);
  if (pat[0]=='?') {
    if (str[0]==0) return 0;
    str++; pat++; goto top;
  }
  if (pat[0]=='*') {
    while (1) {
      if (string_match(str, pat+1)) return 1;
      if (str[0]==0) return 0;
      str++;
    }
  }
  c1 = str[0]; c2 = pat[0];
  if (c1==c2) { str++; pat++; goto top; }
  if ((c1>='a')&&(c1<='z')) c1-=32;
  if ((c2>='a')&&(c2<='z')) c2-=32;
  if (c1==c2) { str++; pat++; goto top; }
  return 0;
}

/*****************************************************************************
 *                                                                           *
 * PATH                                                                      *
 *                                                                           *
 * path_simplify(P)                                                          *
 *                                                                           *
 *  - P is a pointer to a buffer containing a path.  All ".." and "."        *
 *    components of the path are expanded out.                               *
 *                                                                           *
 * path_concat(P, rel)                                                       *
 *                                                                           *
 *  - P is a pointer to a buffer containing a path, rel is another path      *
 *    relative to P.  The logical concatenation of the two paths are         *
 *    stored in the buffer.                                                  *
 *                                                                           *
 * path_absolute(P)                                                          *
 *                                                                           *
 *  - If P is a relative path, it is assumed to be relative to the current   *
 *    directory, and it is thereby converted into an absolute path.          *
 *                                                                           *
 * path_search(name, searchpath)                                             *
 *                                                                           *
 *  - name is a pointer to a buffer containing a program name or program     *
 *    path, and searchpath is a string containing a list of directories      *
 *   (as might be returned by getenv("PATH")).  The program's executable     *
 *    is located and its absolute, readlinked path is stored in the name     *
 *    buffer.                                                                *
 *                                                                           *
 *****************************************************************************/

char *path_segs[100];
int   path_nsegs;

static void path_segs_free()
{
  while (path_nsegs) free(path_segs[--path_nsegs]);
}

static void path_dissect(path)
     char *path;
{
  char buf[1000];
  int len=0;
  while (1) {
    if ((*path=='/')||(*path==0)) {
      buf[len]=0;
      path_segs[path_nsegs++] = strdup(buf);
      len=0;
    }
    else buf[len++] = *path;
    if (*path==0) break;
    path++;
  }
}

static void path_reconstitute(buff)
     char *buff;
{
  int i;
  for (i=0; i<path_nsegs; i++) {
    strcpy(buff, path_segs[i]);
    buff+=strlen(buff);
    *buff++ = '/';
  }
  *(--buff)=0;
}

void path_reduce()
{
  int src, dst;
  src = 0; dst = 0;
  for (src=0; src<path_nsegs; src++) {
    char *t;
    if ((strcmp(path_segs[src],"")==0)&&(src!=0)) continue;
    if (strcmp(path_segs[src],".")==0) continue;
    if (strcmp(path_segs[src],"..")==0) { if (dst) dst--; continue; }
    t = path_segs[dst]; path_segs[dst]=path_segs[src]; path_segs[src]=t;
    dst ++;
  }
  while (src>dst) free(path_segs[--src]);
  path_nsegs = dst;
}

void path_simplify(path)
     char *path;
{
  path_nsegs = 0;
  path_dissect(path);
  path_reduce();
  path_reconstitute(path);
  path_segs_free();
}

void path_concat(base, rel)
     char *base, *rel;
{
  path_nsegs = 0;
  if (rel[0]!='/') path_dissect(base);
  path_dissect(rel);
  path_reduce();
  path_reconstitute(base);
  path_segs_free();
}

void path_absolute(path)
     char *path;
{
  char buff[1024];
  if (path[0]=='/') return;
  getcwd(buff, 1023);
  path_concat(buff, path);
  strcpy(path, buff);
}

int path_exists(path)
     char *path;
{
  struct stat s;
  int ok = stat(path, &s);
  if (ok>=0) return 1;
  return 0;
}

int path_search(prog, path)
     char *prog, *path;
{
  char *end;
  if (strchr(prog,'/')) {
    path_absolute(prog);
    if (path_exists(prog)) return 0;
    prog[0]=0; return -1;
  }
  if ((path)&&(*path)) while (1) {
    char buff[1024];
    int len;
    end = strchr(path, ':');
    if (end==0) { end=path+strlen(path); }
    len = (end - path);
    if (path[0]=='/') {
      memcpy(buff, path, len);
      buff[len]=0;
      path_concat(buff, prog);
      path_absolute(buff);
      if (path_exists(buff)) { strcpy(prog, buff); return 0; }
    }
    if (*end==0) break;
    path=end+1;
  }
  prog[0]=0; errno=ENOENT; return -1;
}

/**************************************************************************
 *
 * Splitting URLs into their components
 *
 **************************************************************************/

char url_kind[100];    int url_kind_origin=0;
char url_user[100];    int url_user_origin=0;
char url_host[100];    int url_host_origin=0;
char url_port[100];    int url_port_origin=0;
char url_path[4096];   int url_path_origin=0;

char *url_getchunk(char *url, char *pre, char *post, char *buf, int *origin)
{
  int len; char *bgn, *end, *skip;
  buf[0]=0; *origin = 0;
  len = strlen(pre);
  if (len && strncmp(url, pre, len)!=0) return url;
  bgn = url + len;
  end = url + len;
  while (1) {
    char c = *end;
    if ((c==0)||(c==':')||(c=='/')||(c=='@')) break;
    end++;
  }
  len = strlen(post);
  if (len && strncmp(end, post, len)!=0) return url;
  skip = end+len;
  len = end-bgn; if (len>99) len=99;
  memcpy(buf, bgn, len); buf[len]=0;
  *origin = 's';
  return skip;
}

void url_split(char *url)
{
  char *p;
  url = url_getchunk(url, "", "://", url_kind, &url_kind_origin);
  url = url_getchunk(url, "", "@"  , url_user, &url_user_origin);
  url = url_getchunk(url, "", ""   , url_host, &url_host_origin);
  if (url_kind[0]==0) {
    strcpy(url_kind,"RCP");
    url_port[0]=0; url_port_origin=0;
    if (url[0]==0) { url_path[0]=0; url_path_origin=0; return; }
    if (url[0]!=':') goto abort;
    strcpy(url_path, url+1); url_path_origin='s';
    return;
  } else {
    url = url_getchunk(url, ":", ""  , url_port, &url_port_origin);
    if (url[0]==0) { url_path[0]=0; url_path_origin=0; return; }
    if (url[0]!='/') goto abort;
    strcpy(url_path, url); url_path_origin='s';
    return;
  }
 abort:
  fprintf(stderr,"syntax error in resource locator.\n");
}

void url_relative(char *url, char *rel)
{
  char *p;
  rel = url_getchunk(rel, "", "://", url_kind, &url_kind_origin);
  if (url_kind_origin==0) {
    url = url_getchunk(url, "", "://", url_kind, &url_kind_origin);
    url = url_getchunk(url, "", "@"  , url_user, &url_user_origin);
    url = url_getchunk(url, "", ""   , url_host, &url_host_origin);
    url = url_getchunk(url, ":", ""  , url_port, &url_port_origin);
    if (rel[0]=='/') { strcpy(url_path, rel); return; }
    if (url[0]=='/') { strcpy(url_path, url); }
    else { url_path[0]='/'; strcpy(url_path+1, url); }
    p = url_path + strlen(url_path);
    while (*p!='/') p--;
    while (1) {
      if (strncmp(rel,"./",2)==0) { rel+=2; }
      else if (strncmp(rel,"../",3)==0) {
	rel+=3; 
	if (p>url_path) { p--; while (*p!='/') p--; }
      }
      else break;
    }
    p++; 
    while (rel[0]=='/') rel++;
    strcpy(p, rel); url_path_origin='s';
  } else {
    rel = url_getchunk(rel, "", "@"  , url_user, &url_user_origin);
    rel = url_getchunk(rel, "", ""   , url_host, &url_host_origin);
    rel = url_getchunk(rel, ":", ""  , url_port, &url_port_origin);
    if (rel[0]=='/') { strcpy(url_path, rel); }
    else { url_path[0]='/'; strcpy(url_path+1, rel); }
    url_path_origin='s';
  }
}

/**************************************************************************
 *
 * SKT - socket routines
 *
 * void skt_server(unsigned int *pip, unsigned int *ppo, unsigned int *pfd)
 *
 *   - create a tcp server socket.  Performs the whole socket/bind/listen
 *     procedure.  Returns the IP address of the socket (eg, the IP of the
 *     current machine), the port of the socket, and the file descriptor.
 *
 * void skt_accept(int src,
 *                 unsigned int *pip, unsigned int *ppo, unsigned int *pfd)
 *
 *   - accepts a connection to the specified socket.  Returns the
 *     IP of the caller, the port number of the caller, and the file
 *     descriptor to talk to the caller.
 *
 * int skt_connect(unsigned int ip, int port)
 *
 *   - Opens a connection to the specified server.  If blocking is set,
 *     then skt_connect waits for the socket to finish connecting.
 *
 **************************************************************************/

int skt_server(int desiredport, int *pip, int *ppo, int *pfd)
{
  int fd= -1;
  int ok, len;
  struct sockaddr_in addr;
  
  fd = socket(PF_INET, SOCK_STREAM, 0);
  if (fd < 0) return -1;
  memset(&addr, 0, sizeof(addr));
  addr.sin_family = AF_INET;
  addr.sin_port = desiredport;
  ok = bind(fd, (struct sockaddr *)&addr, sizeof(addr));
  if (ok < 0) { close(fd); return -1; }
  ok = listen(fd,5);
  if (ok < 0) { close(fd); return -1; }
  len = sizeof(addr);
  ok = getsockname(fd, (struct sockaddr *)&addr, &len);
  if (ok < 0) { close(fd); return -1; }

  *pfd = fd;
  *pip = ip_lookup(0);
  *ppo = ntohs(addr.sin_port);
  return 0;
}

int skt_accept(int src,
		unsigned int *pip, unsigned int *ppo, unsigned int *pfd)
{
  int i, fd, ok;
  struct sockaddr_in remote;
  i = sizeof(remote);
  fd = accept(src, (struct sockaddr *)&remote, &i);
  if (fd<0) return fd;
  *pip=htonl(remote.sin_addr.s_addr);
  *ppo=htons(remote.sin_port);
  *pfd=fd;
  return 0;
}

int skt_connect(unsigned int ip, int port)
{
  struct sockaddr_in remote; short sport=port;
  int fd, ok, len;
    
  /* create an address structure for the server */
  memset(&remote, 0, sizeof(remote));
  remote.sin_family = AF_INET;
  remote.sin_port = htons(sport);
  remote.sin_addr.s_addr = htonl(ip);
    
  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0) { return -1; }
  
 conn:
  ok = connect(fd, (struct sockaddr *)&(remote), sizeof(remote));
  if (ok<0) { int i=errno; close(fd); errno=i; return -1; }
  return fd;
}

/*****************************************************************************
 *
 * IP address lookup
 *
 *****************************************************************************/

int ip_lookup(host)
     char *host;
{
  static unsigned int myip = 0;   int nread, ip;
  unsigned int usecount, ip0, ip1, ip2, ip3;
  struct hostent *hostent; char hostname[100];

  /* Strategy One - if host is null, return MY ip address */
  if ((host==0)||(host[0]==0)) {
    if (myip != 0) return myip;
    if (gethostname(hostname, 99)<0) { perror("gethostname"); exit(1); }
    hostent = gethostbyname(hostname);
    if (hostent == 0) { perror("gethostbyname"); exit(1); }
    myip = htonl(*((int *)(hostent->h_addr_list[0])));
    return myip;
  }
  
  /* Strategy Two - if host is numeric, just parse it */
  nread = sscanf(host, "%d.%d.%d.%d", &ip0,&ip1,&ip2,&ip3);
  if (nread==4) return ((ip0<<24)+(ip1<<16)+(ip2<<8)+ip3);
  
  /* Strategy Three: Go to DNS nameserver */
  hostent = gethostbyname(host);
  if (hostent == 0) return 0;
  ip = htonl(*((int *)(hostent->h_addr_list[0])));
  return ip;
}

int ip_elookup(char *host)
{
  int ip = ip_lookup(host);
  if (ip==0) {
    fprintf(stderr,"Unknown host %s\n",host);
    exit(1);
  }
  return ip;
}

char *ip_string(unsigned int ip)
{
  static char buffer[20];
  sprintf(buffer, 
	  "%d.%d.%d.%d",(ip>>24),(ip>>16)&0xFF,(ip>>8)&0xFF,ip&0xFF);
  return buffer;
}

/******************************************************************************
 *
 * Basic operations
 *
 *****************************************************************************/

void u_add_line(int argc, char **argv)
{
  FILE *f; char line[1024]; int len;
  if (argc!=2) usage();
  len = strlen(argv[1]);
  f = fopen(argv[0], "r");
  if (f) {
    while (fgets(line, 1023, f)) {
      if ((strlen(line)==len+1)&&(strncmp(line,argv[1],len)==0))
	exit(0);
    }
    fclose(f);
  }
  f = efopen(argv[0], "a");
  fprintf(f,"%s\n",argv[1]);
  fclose(f);
  exit(0);
}

void u_claim_lockfile(int argc, char **argv)
{
  int fd;
  if (argc != 1) usage();
  while (1) {
    fd = open(argv[0], O_RDWR | O_CREAT | O_EXCL, 0777);
    if (fd >= 0) break;
    if (errno != EEXIST) {
      fprintf(stderr,"claim-lockfile: %s\n", strerror(errno));
      exit(1);
    }
    sleep(1);
  }
  exit(0);
}

void u_fetch_url(int argc, char **argv)
{
  char *headfn = 0; FILE *headf=0;
  char *datafn = 0; FILE *dataf=0;
  char *linkfn = 0; FILE *linkf=0;
  char line[10000]; char *url, *p, *e; FILE *sktw, *sktr;
  char *keys[100], *vals[100]; int minsize=0, totsize=0, timeout=60, nkv;
  unsigned int ip, port; int c, state, taglen, skt, i, ishtml;

  while (argv[0]&&(argv[0][0]=='-')) {
    if (argv[1]==0) usage();
    if      (strcmp(*argv,"-u")==0) { argv++; url= *argv++; }
    else if (strcmp(*argv,"-h")==0) { argv++; headfn= *argv++; }
    else if (strcmp(*argv,"-l")==0) { argv++; linkfn= *argv++; }
    else if (strcmp(*argv,"-d")==0) { argv++; datafn= *argv++; }
    else if (strcmp(*argv,"-t")==0) { argv++; timeout=atoi(*argv); argv++; }
    else if (strcmp(*argv,"-m")==0) { argv++; minsize=atoi(*argv); argv++; }
    else usage();
  }
  if (argv[0]) usage();
  
  if (url==0) {
    fprintf(stderr,"Must specify -u URL\n");
    exit(1);
  }
  
  if (headfn) headf=tfopen(headfn);
  if (datafn) dataf=tfopen(datafn);
  if (linkfn) linkf=tfopen(linkfn);
  
  url_split(url);
  if ((url_kind_origin != 's')||(strcasecmp(url_kind,"http"))) {
    fprintf(stderr,"Must specify an http URL.\n");
    exit(1);
  }
  if (url_user[0]) {
    fprintf(stderr,"May not specify a username with http protocol.\n");
    exit(1);
  }
  
  ip = ip_elookup(url_host);
  if (url_port[0]==0) port = 80;
  else port = atoi(url_port);
  skt = skt_connect(ip, port);
  if (skt < 0) {
    fprintf(stderr,"connect: %s\n", strerror(errno));
    exit(1);
  }
  
  sktr = fdopen(skt, "r");
  sktw = fdopen(skt, "w");
  p = strchr(url_path, '#'); if (p) *p=0;
  fprintf(sktw, "GET %s HTTP/1.0\r\n", url_path);
  fprintf(sktw, "User-Agent: jnet\r\n");
  fprintf(sktw, "Host: %s:%s\n", url_host, url_port);
  fprintf(sktw, "Accept: */*\r\n");
  fprintf(sktw, "\r\n");
  fflush(sktw);
  ishtml = 0;
  while (1) {
    fgets(line, 9999, sktr);
    p = line; while ((*p==' ')||(*p=='\t')) p++;
    if ((*p=='\r')||(*p=='\n')||(*p==0)) break;
    e = p; while ((*e)&&(*e!='\r')&&(*e!='\n')) e++;
    *e = 0;
    if (headf) fprintf(headf, "%s\n", p);
    if (strncasecmp(p,"Content-type:",13)==0) {
      p+=13; while ((*p==' ')||(*p=='\t')) p++;
      if (strncasecmp(p,"text/html",9)==0) ishtml=1;
    }
  }
  if (headf) fprintf(headf,"Request: %s\n", url);
  if ((!ishtml)||(linkfn==0)) {
    if (dataf) 
      while ((c=getc(sktr))!=EOF) { putc(c, dataf); totsize++; }
  } else {
    state=0;
    while ((c=getc(sktr))!=EOF) {
      if (dataf) { putc(c, dataf); totsize++; }
    dispatch:
      switch (state) {
      case 0:
	if (c=='<') { taglen=0; state=1; nkv=0; }
	break;
      case 1:
	if ((c==' ')||(c=='\t')||(c=='\r')||(c=='\n')) break;
	state=2;
      case 2:
	if (c=='>') { state=10; goto dispatch; }
	state=3;
      case 3:
	if ((c>='a')&&(c<='z')) { line[taglen++]=c-32; break; }
	if ((c>='A')&&(c<='Z')) { line[taglen++]=c; break; }
	if ((c>='0')&&(c<='9')) { line[taglen++]=c; break; }
	if (taglen==0) { state=0; break; }
	line[taglen]=0; keys[nkv] = strdup(line); taglen=0;
	state=4;
      case 4:
	if ((c==' ')||(c=='\t')||(c=='\r')||(c=='\n')) break;
	state=5;
      case 5:
	if (c=='=') { state=6; break; }
	vals[nkv]="+"; nkv++; 
	if (c=='>') { state=10; goto dispatch; }
	state=3; goto dispatch;
      case 6:
	if ((c==' ')||(c=='\t')||(c=='\r')||(c=='\n')) break;
	state=7;
      case 7:
	if (c!='"') { state=9; goto dispatch; }
	taglen=0; state=8; break;
      case 8:
	if (c!='"') { line[taglen++] = c; break; }
	line[taglen]=0; vals[nkv]=strdup(line); taglen=0; nkv++;
	state=1; break;
      case 9:
	if ((c!=' ')&&(c!='\t')&&(c!='\r')&&(c!='\n')&&(c!='>'))
	  { line[taglen++] = c; break; }
	line[taglen]=0; vals[nkv]=strdup(line); taglen=0; nkv++;
	state=1; goto dispatch;
      case 10:
	if (nkv) {
	  for(i=1; i<nkv; i++) 
	    if ((strcmp(keys[i],"HREF")==0)||(strcmp(keys[i],"SRC")==0)) {
	      url_relative(url, vals[i]);
	      fprintf(linkf,"%s://%s", url_kind, url_host);
	      if (url_port[0]) fprintf(linkf, ":%s", url_port);
	      fprintf(linkf,"%s\n", url_path);
	    }
	}
	state=0; break;
      }	
    }
  }
  fclose(sktr); fclose(sktw); close(skt);
  tfclose(dataf, datafn);
  tfclose(headf, headfn);
  tfclose(linkf, linkfn);
  if ((datafn)&&(totsize<minsize)) unlink(datafn);
}

void u_for(int argc, char **argv)
{
  int lo, hi, i;
  if ((argc < 2)||(argc > 4)) usage();
  lo = atoi(argv[0]);
  hi = atoi(argv[1]);
  for (i=lo; i<=hi; i++) {
    if (argc > 2) printf("%s", argv[2]);
    printf("%d",i);
    if (argc > 3) printf("%s", argv[3]);
    printf("\n");
  }
}

void u_getpass(int argc, char **argv)
{
  if (argc!=1) usage();
  printf("%s\n", getpass(argv[0]));
  exit(0);
}

void u_ip(int argc, char **argv)
{
  unsigned int ip;
  if (argc != 1) usage();
  ip = ip_elookup(argv[0]);
  printf("%s\n", ip_string(ip));
}

void u_local_ip(int argc, char **argv)
{
  unsigned int ip = ip_elookup(0);
  if (argc != 0) usage();
  printf("%s\n", ip_string(ip));
}

void u_net_connect(int argc, char **argv)
{
  int file, port, nread, ip;
  if (argc != 2) usage();
  ip = ip_elookup(argv[0]);
  port = atoi(argv[1]);
  if (port==0) {
    fprintf(stderr,"port number expected.\n");
    exit(1);
  }
  file = skt_connect(ip, port);
  if (file<0) { perror("skt_connect"); exit(1); }
  relay(0, file, file, 1);
}

void u_net_server(int argc, char **argv)
{
  int port, nread, ok;
  int m_ip, m_port, m_skt;
  int c_ip, c_port, c_skt;
  if (argc != 1) usage();
  nread = sscanf(argv[0], "%d", &port);
  if (nread != 1) usage();
  ok = skt_server(port, &m_ip, &m_port, &m_skt);
  if (ok<0) { perror("creating server socket"); exit(1); }
  if (m_port != port) { fprintf(stderr,"barfola.\n"); exit(1); }
  ok = skt_accept(m_skt, &c_ip, &c_port, &c_skt);
  if (ok<0) { perror("accepting remote connection"); exit(1); }
  shutdown(m_skt, 2); close(m_skt);
  relay(0, c_skt, c_skt, 1);
}

void u_net_repeater(int argc, char **argv)
{
  int localport, remhost, remport, ok, r_skt, i, once;
  int m_ip, m_port, m_skt;
  int c_ip, c_port, c_skt;
  once = 0;
  if (strcmp(argv[0], "--once")==0) { once=1; argv++; argc--; }
  if (argc != 3) usage();
  localport = atoi(argv[0]);
  remhost = ip_elookup(argv[1]);
  remport = atoi(argv[2]);
  if ((localport==0)||(remport==0)) usage();
  
  ok = skt_server(localport, &m_ip, &m_port, &m_skt);
  if ((ok<0)&&(errno==EADDRINUSE)) exit(0);
  if (ok<0) {
    fprintf(stderr,"creating server socket: %s\n", strerror(errno));
    exit(1);
  }
  if (fork()==0) {
    int i; for (i=0; i<256; i++) { if (i!=m_skt) close(i); }
    while (1) {
      ok = skt_accept(m_skt, &c_ip, &c_port, &c_skt);
      if (ok<0) exit(1);
      if (fork()==0) {
	close(m_skt);
	r_skt = skt_connect(remhost, remport);
	if (r_skt < 0) exit(0);
	relay(c_skt, r_skt, r_skt, c_skt);
      }
      close(c_skt);
      if (once) exit(0);
    }
  }
}

void u_net_check_server(int argc, char **argv)
{
  int host, port, timeout, begintime, skt;
  if (argc != 3) usage();
  host = ip_elookup(argv[0]);
  port = atoi(argv[1]);
  timeout = atoi(argv[2]);
  
  begintime = time(0);
  while (time(0) < begintime + timeout) {
    skt = skt_connect(host, port);
    if (skt >= 0) exit(0);
    sleep(1);
  }
  fprintf(stderr,"Cannot connect to server.\n");
  exit(1);
}

void u_older_days(int argc, char **argv)
{
  int days, secs, t, now;
  if (argc != 2) usage();
  days = atoi(argv[0]);
  secs = days * 24 * 60 * 60;
  t = filetime(argv[1]);
  now = time(0);
  if (t < (now-secs)) exit(0);
  exit(1);
}

void u_older_seconds(int argc, char **argv)
{
  int secs, t, now;
  if (argc != 2) usage();
  secs = atoi(argv[0]);
  t = filetime(argv[1]);
  now = time(0);
  if (t < (now-secs)) exit(0);
  exit(1);
}

void u_older_than(int argc, char **argv)
{
  int t1, t2;
  if (argc != 2) usage();
  t1 = filetime(argv[0]);
  t2 = filetime(argv[1]);
  if (t1 < t2) exit(0);
  exit(1);
}

void u_path_relative_to(int argc, char **argv)
{
  char path[MAXPATHLEN];
  struct stat preinfo;
  struct stat pathinfo;
  char *pre, *p; int len, ok;
  
  /* Do the string-manipulation test first */
  if (argc!=2) usage();
  strcpy(path, argv[1]);
  path_absolute(path);
  pre = getenv(argv[0]);
  if (pre == 0) { printf("%s\n", path); exit(0); }
  len = strlen(pre);
  if ((len>0)&&(pre[len-1]=='/')) pre[--len] = 0;
  if ((strlen(path)>len) && (path[len]=='/') && (strncmp(path,pre,len)==0)) {
    printf("$%s/%s\n", argv[0], path+len+1); exit(0);
  }

  /* if that fails, try the file-manipulation test */
  ok = stat(pre, &preinfo);
  if (ok<0) { printf("%s\n", path); exit(0); }
  for (p=path; ; p++) {
    int ch = *p;
    if ((ch=='/')||(ch==0)) {
      *p = 0;
      ok = stat(rootify(path), &pathinfo);
      if (ok<0) { *p=ch; printf("%s\n", path); exit(0); }
      if ((pathinfo.st_ino == preinfo.st_ino)&&
          (pathinfo.st_dev == preinfo.st_dev)) {
	if (ch) printf("$%s/%s\n", argv[0], p+1);
	else printf("$%s\n", argv[0]);
	exit(0);
      }
      *p = ch;
    }
    if (ch==0) break;
  }
  printf("%s\n", path); exit(0);
}

void u_path_simplify(int argc, char **argv)
{
}

void u_path_standardize(int argc, char **argv)
{
}

void u_random_line(int argc, char **argv)
{
  struct stat info; int bufsize, fd, ok, i, nlines, line; char *buf, *p;
  if (argc != 1) usage();
  ok = stat(argv[0], &info);
  if (ok<0) { perror("random-line: stat"); exit(1); }
  bufsize = info.st_size+2;
  buf = malloc(bufsize);
  if (buf==0) { 
    fprintf(stderr,"random-line: memory allocation failed.\n");
    exit(1);
  }
  buf[0]='\n';
  buf[bufsize-1] = '\n';
  fd = open(argv[0], O_RDONLY, 0);
  if (fd < 0) { perror("random-line: open"); exit(1); }
  ok = read(fd, buf+1, info.st_size);
  if (ok < 0) { perror("random-line: read"); exit(1); }
  if (ok != info.st_size) {
    fprintf(stderr,"random-line: read failed.\n");
    exit(1);
  }
  nlines = 0;
  for (p=buf; p<buf+bufsize-1; p++)
    if ((p[0]=='\n')&&(p[1]!='\n'))
      nlines++;
  if (nlines==0) {
    fprintf(stderr,"random-line: file is empty.\n");
    exit(1);
  }
  jseed();
  line = (jrand()&0x7fffffff) % nlines;
  nlines=0; p=buf;
  while (nlines<=line) {
    if ((p[0]=='\n')&&(p[1]!='\n'))
      nlines++;
    p++;
  }
  while (*p&&(*p!='\n')) { putchar(*p); p++; }
  putchar('\n');
}

void u_random_number(int argc, char **argv, int mod)
{
  char *digits="0123456789ABCDEF";
  int i, len;
  if (argc!=1) usage();
  len = atoi(argv[0]);
  if ((len<1)||(len>16384)) usage();
  jseed();
  for(i=0; i<len; i++)
    {
      int n=((jrand()&0x7FFFFFFF)%mod);
      putchar(digits[n]);
    }
  putchar('\n');
}

void u_random_pick(int argc, char **argv)
{
  int pick;
  if (argc<1) usage();
  jseed();
  pick = (jrand()&0x7fffffff)%argc;
  printf("%s\n", argv[pick]);
}

void u_random_sort(int argc, char **argv)
{
  char line[10000];
  char **lines; int max, total, len, n;
  if (argc!=0) usage();
  lines = malloc(100*sizeof(char*));
  max = 100; total=0;
  while (fgets(line, 9999, stdin)) {
    len = strlen(line);
    while ((len>0)&&(line[len-1]<=32)) len--;
    line[len] = 0;
    if (total==max) {
      max<<=1;
      lines = realloc(lines, max*sizeof(char*));
    }
    lines[total++] = strdup(line);
  }
  jseed();
  if (total) while (1) {
    n = (jrand()&0x7FFFFFFF)%total;
    printf("%s\n", lines[n]);
    total--;
    if (total==0) break;
    lines[n] = lines[total];
  }
}

void u_search_path(int argc, char **argv)
{	
  int i;
  char buff[8192];
  char *path=getenv("PATH");
  for (i=0; i<argc; i++)
    {
      strcpy(buff, argv[i]);
      if (path_search(buff, path)==0)
	{ printf("%s\n",buff); exit(0); }
    }
  exit(1);
}

void u_strip_ctrl_chars(int argc, char **argv)
{
  int c;
  if (argc!=0) usage();
  while ((c=getchar())!=EOF)
    if (((c>=32)&&(c<127))||(c==10)||(c==9)) putchar(c);
  exit(0);
}

void u_wait_file(argc, argv)
     int argc; char **argv;
{
  int ok1; struct stat stat1;
  int ok2; struct stat stat2;
  if (argc != 1) usage();
  memset(&stat1, 0, sizeof(stat1));
  memset(&stat2, 0, sizeof(stat2));
  ok1 = stat(argv[0], &stat1);
  while (1) {
    ok2 = stat(argv[0], &stat2);
    if ((ok1 != ok2) || (stat1.st_mtime != stat2.st_mtime))
      exit(0);
    sleep(1);
  }
}

void usage()
{
#define e(s) fprintf(stderr,s)
e("jnet usages:\n");
e("\n");
e("  jnet help\n");
e("  jnet claim-lockfile [--warn seconds] filename\n");
e("  jnet fetch-url -u URL [-l links] [-d data] [-h headers] [-t timeout] [-m minsize]\n");
e("  jnet for lo hi\n");
e("  jnet getpass prompt\n");
e("  jnet ip hostname\n");
e("  jnet local-ip\n");
e("  jnet net-connect <host> <port>\n");
e("  jnet net-server <port>\n");
e("  jnet net-repeater [--once] <localport> <remotehost> <remoteport>\n");
e("  jnet net-check-server <host> <port> <timeout>\n");
e("  jnet older-days seconds filename\n");
e("  jnet older-seconds days filename\n");
e("  jnet older-than file1 file2\n");
e("  jnet path-simplify path\n");
e("  jnet path-standardize path\n");
e("  jnet path-relative-to varname path\n"); 
e("  jnet random-decimal numdigits\n");
e("  jnet random-hex numdigits\n");
e("  jnet random-line file\n");
e("  jnet random-pick choice1 choice2 ...\n");
e("  jnet random-sort < file1 > file2\n");
e("  jnet search-path prog1 prog2 prog3...\n");
e("  jnet strip-ctrl-chars < file1 > file2\n");
e("  jnet wait-file filename\n");
e("\n");
#undef e
exit(1);
}

main(int argc, char **argv)
{
  char *p;
  int len = strlen(argv[0]);
  p = (char *)strrchr(argv[0],'/');
  if (p) argv[0]=(p+1);
  if (argc<1) usage();
  if (strcmp(argv[0],"jnet")==0) { argc--; argv++; }
  if (argc<1) usage();
  p = argv[0];
  if (eqs(p,"add-line"))         {u_add_line(argc-1, argv+1); exit(0); }
  if (eqs(p,"claim-lockfile"))   {u_claim_lockfile(argc-1, argv+1); exit(0); }
  if (eqs(p,"fetch-url"))        {u_fetch_url(argc-1, argv+1); exit(0); }
  if (eqs(p,"for"))              {u_for(argc-1, argv+1); exit(0); }
  if (eqs(p,"getpass"))          {u_getpass(argc-1, argv+1); exit(0); }
  if (eqs(p,"ip"))               {u_ip(argc-1, argv+1); exit(0); }
  if (eqs(p,"local-ip"))         {u_local_ip(argc-1, argv+1); exit(0); }
  if (eqs(p,"net-connect"))      {u_net_connect(argc-1, argv+1); exit(0); }
  if (eqs(p,"net-server"))       {u_net_server(argc-1, argv+1); exit(0); }
  if (eqs(p,"net-repeater"))     {u_net_repeater(argc-1, argv+1); exit(0); }
  if (eqs(p,"net-check-server")) {u_net_check_server(argc-1, argv+1); exit(0);}
  if (eqs(p,"older-days"))       {u_older_days(argc-1, argv+1); exit(0); }
  if (eqs(p,"older-seconds"))    {u_older_seconds(argc-1, argv+1); exit(0); }
  if (eqs(p,"older-than"))       {u_older_than(argc-1, argv+1); exit(0); }
  if (eqs(p,"path-simplify"))    {u_path_simplify(argc-1, argv+1);exit(0);}
  if (eqs(p,"path-standardize")) {u_path_standardize(argc-1, argv+1); exit(0);}
  if (eqs(p,"path-relative-to")) {u_path_relative_to(argc-1, argv+1); exit(0);}
  if (eqs(p,"random-decimal"))   {u_random_number(argc-1,argv+1,10);exit(0); }
  if (eqs(p,"random-hex"))       {u_random_number(argc-1,argv+1,16);exit(0); }
  if (eqs(p,"random-line"))      {u_random_line(argc-1,argv+1);exit(0);}
  if (eqs(p,"random-pick"))      {u_random_pick(argc-1,argv+1);exit(0);}
  if (eqs(p,"random-sort"))      {u_random_sort(argc-1,argv+1);exit(0); }
  if (eqs(p,"search-path"))      {u_search_path(argc-1, argv+1); exit(0); }
  if (eqs(p,"strip-ctrl-chars")) {u_strip_ctrl_chars(argc-1, argv+1); exit(0);}
  if (eqs(p,"wait-file"))        {u_wait_file(argc-1, argv+1); exit(0); }
  usage();
}

